#
# Edition operations
#
######################################################################
# Empty_Graph -
# Procedure to clean the piece editor after asking for confirmation.
#
# Parameters:
# - graph	the canvas where this happens.
#
proc Empty_Graph { graph } {
    global cur_piece state

    if { $state(.graph) != "saved" } {
	switch [tk_dialog .quit "Modified" "The piece has been modified. Do you wish to store it?" {warning} 0 {Store (current) & New} {Store (new) & New} {New} {Cancel}] {
	    0 { 
		Store_Current_Piece 0 1
	    }
	    1 { 
		Store_Current_Piece 1 1
	    }
	    2 { }
	    3 { return }
	}
    }

    if [winfo exists $graph] {
	$graph.canvas delete PIECE HIGHLIGHT
    }

    set cur_piece {}
    set state(.graph) saved
}

######################################################################
# Clear_Store -
# Procedure that clears an area of the Store display between two levels
# in order to prepare for redrawing. 
# Not to be confused with "Erase_Store" which deletes pieces.
#
# Parameters:
# - from	the number of the level from which to clear (inclusive)
# - to		the number of the level to which to clear (exclusive)
#
proc Clear_Store { from to } {
    global sq_size nbrows
    global DEBUG

    if { 0 == [string compare $to end] } {
	set yfrom 0
    } else {
	set yfrom [expr ($nbrows(.store)-$to)*$sq_size(.store)]
    }
    set yto   [expr ($nbrows(.store)-$from)*$sq_size(.store)]

    if $DEBUG {
	puts "deleting store in range \[$from,$to\] -> \[$yfrom,$yto\["
    }

    foreach type { SEP PIECE HIGHLIGHT TEXT } {

	set trucs [.store.canvas find withtag $type]
	if $DEBUG {
	    puts "$type = $trucs"
	}
	foreach obj $trucs { 
	    set ypos [lindex [.store.canvas coords $obj] 1]
	    if [expr { $ypos >= $yfrom } && { $ypos < $yto }] {
		.store.canvas delete $obj
	    }
	}
    }

}

######################################################################
# Empty_Store -
# Procedure to clean the store after asking for confirmation.
#
# Parameters: none
#
proc Empty_Store { } {
    global piece_list cur_piece_nb state

    if { $state(.store) != "saved" } {
	switch [tk_dialog .quit "Modified" "The store has been modified. Do you wish to save it?" {warning} 0 {Save & Clear} {Clear, don't save} {Cancel}] {
	    0 {
		if ![Save_Tetris] { return }
	    }
	    1 { }
	    2 { return }
	}
    }

    Clear_Store 0 end

    set piece_list {}
    set cur_piece_nb 0

    set state(.store) saved
}

proc Make_Canvas { { new_square_size 0 } } {
    global sq_size nbrows square_size store_ratio cur_width pile_height
    global view_range state
    global DEBUG

    # puts "Making a canvas with sqsize $new_square_size"

    if { $new_square_size > 0 } { 
	set square_size $new_square_size 
    }

    set sq_size(.piece_edit.graph) $square_size
    set sq_size(.graph) $square_size
    set sq_size(.store) [expr $sq_size(.graph) / $store_ratio]
    set sq_size(.drop)  $sq_size(.graph)

    set nbcols $cur_width
    set nbrows(.piece_edit.graph) [expr $nbcols*4]
    set nbrows(.graph) [expr $nbcols*4]
    set nbrows(.store) [expr $nbcols*10]
    set nbrows(.drop)  [expr $nbcols*20 < 500 ? 500 : $nbcols*20 ]

    # adjusting width if too large
    if [expr $sq_size(.drop)*$cur_width > 500] {
	set sq_size(.drop) [expr 500 / $cur_width]
	if [expr $sq_size(.drop) < 2] {
	    set sq_size(.drop) 2
	}
    }
    # making the edit frame (.graph)
    set width [expr $nbcols*$sq_size(.graph)]
    set height [expr $nbrows(.graph)*$sq_size(.graph)]
    set visible_height [expr $height>600?600:$height]

    if { $DEBUG } {
	puts "VH = $visible_height"
    }

    if [winfo exists .graph] {
	.graph.canvas configure -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height]
    } else {
	frame .graph
	canvas .graph.canvas -background lightblue \
	    -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height] \
	    -yscrollcommand [list .graph.yscroll set]
	scrollbar .graph.yscroll -orient vertical -width 10 \
	    -command [list .graph.canvas yview]
    }

    # geometry of dropzone is different
    set width [expr $nbcols*$sq_size(.drop)]
    set height [expr $nbrows(.drop)*$sq_size(.drop)]
    set visible_height [expr $height>600?600:$height]

    if [winfo exists .drop] {
	.drop.canvas configure -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height]
    } else {
	frame .drop
	menubutton .drop.label -text "Drop Zone" \
	   -menu .drop.label.menu -underline 0
	set m [menu .drop.label.menu]
	$m add command -label "Clear" -underline 0 -command {Clear_Drop}
	pack .drop.label -side top -anchor c

	canvas .drop.canvas -background lightyellow \
	    -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height] \
	    -yscrollcommand [list .drop.yscroll set]
	scrollbar .drop.yscroll -orient vertical -width 10 \
	    -command [list .drop.canvas yview]
    }
    set view_range(low) 0
    set view_range(up)  [expr double($visible_height)/($sq_size(.drop)*double($nbrows(.drop)))]

    # geometry of store also different
    # Width contains two extra columns for number info.
    set width [expr (2+$nbcols)*$sq_size(.store)]
    set height [expr $nbrows(.store)*$sq_size(.store)]
    set visible_height [expr $height>600?600:$height]

    if [winfo exists .store] {
	.store.canvas configure -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height]
    } else {
	frame .store
	menubutton .store.label -text "Piece Store" \
	    -menu .store.label.menu -underline 0 
	set m [menu .store.label.menu]
	$m add command -label "Clear" -underline 0 -command {Empty_Store}
	$m add command -label "Redraw" -underline 0 \
	    -command "Clear_Store 0 end\nDisplay_Store 1 end"
	$m add command -label "Set Size" -underline 0 \
	    -command {Make_Store_Editor}
	pack .store.label -side top -anchor c

	canvas .store.canvas -background salmon \
	    -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height] \
	    -yscrollcommand [list .store.yscroll set]
	scrollbar .store.yscroll -orient vertical -width 10 \
	    -command [list .store.canvas yview]
    }

    pack .store.canvas -side left
    pack .store.yscroll -side right -fill y
    pack .store -side right -anchor s

#     pack .graph.canvas -side left -fill y -expand true
#     pack .graph.yscroll -side right -fill y
#     pack .graph -side right

    pack .drop.label -side top
    pack .drop.canvas -side left -fill y
    pack .drop.yscroll -side right -fill y
    pack .drop -side right -fill y -expand true

    .graph.canvas yview moveto 1.0
    .store.canvas yview moveto 1.0
    .drop.canvas yview moveto 1.0

    # draw the grid
    for { set i 0 } { $i <= $nbcols } { incr i 1 } {
	set x [expr $i*$sq_size(.graph)]
	.graph.canvas create line $x 0 $x [expr $nbrows(.graph)*$sq_size(.graph)] \
	    -tag {SEP HORR}
    }
    for { set i 0 } { $i <= $nbrows(.graph) } { incr i 1 } {
	set y [expr $i*$sq_size(.graph)]
	.graph.canvas create line 0 $y [expr $nbcols*$sq_size(.graph)] $y \
	    -tag {SEP HORR}
    }

    set state(.store) saved
    set state(.drop)  saved
    set state(.graph) saved

    Clear_Drop

    Reset_Standard_Bindings

}

proc Make_Edit_Window { } {
    global sq_size nbrows cur_width img_dir DX DY
    global default_piece_props cur_piece_props
    global DEBUG

    if [catch {toplevel .piece_edit}] { return }

    wm title .piece_edit "Tetris Piece Editor"

    # create the top menu
    frame .piece_edit.topmenu -relief raised -class MenFrame -borderwidth 2 
    pack .piece_edit.topmenu -side top -fill x

    menubutton .piece_edit.topmenu.file -text "Store..." \
	-menu .piece_edit.topmenu.file.menu \
	-underline 0
    set m [menu .piece_edit.topmenu.file.menu -tearoff 1]
    $m add command -label "Store as Current" -underline 9 \
	-command "Store_Current_Piece 0 0"
    $m add command -label "Store as New" -underline 9 \
	-command "Store_Current_Piece 1 0"
    $m add command -label "Store and New" -underline 0 \
	-command "Store_Current_Piece 0 1"
    $m add command -label "Clear" -underline 0 \
	-command "Empty_Graph .piece_edit.graph"
    $m add separator
    $m add command -label "Dismiss" -underline 0 \
	-command "Quit_Piece_Editor"
    pack .piece_edit.topmenu.file -side left

    menubutton .piece_edit.topmenu.move -text "Move" \
	-menu .piece_edit.topmenu.move.menu \
	-underline 0
    set m [menu .piece_edit.topmenu.move.menu -tearoff 1]
    $m add command -label "Up & Left"    -command { Move_Piece -1  1 }
    $m add command -label "Up"           -command { Move_Piece  0  1 }
    $m add command -label "Up & Right"   -command { Move_Piece  1  1 }
    $m add command -label "Left"         -command { Move_Piece -1  0 }
    $m add command -label "Right"        -command { Move_Piece  1  0 }
    $m add command -label "Down & Left"  -command { Move_Piece -1 -1 }
    $m add command -label "Down"         -command { Move_Piece  0 -1 }
    $m add command -label "Down & Right" -command { Move_Piece  1 -1 }
    pack .piece_edit.topmenu.move -side left

    menubutton .piece_edit.topmenu.rotate -text "Rotate" \
	-menu .piece_edit.topmenu.rotate.menu \
	-underline 0
    set m [menu .piece_edit.topmenu.rotate.menu -tearoff 1]
    $m add command -label "Clockwise"    -command { Rotate_Piece  1 }
    $m add command -label "Counter-clockwise" -command { Rotate_Piece  3 }
    pack .piece_edit.topmenu.rotate -side left

    menubutton .piece_edit.topmenu.mirror -text "Mirror" \
	-menu .piece_edit.topmenu.mirror.menu \
	-underline 0
    set m [menu .piece_edit.topmenu.mirror.menu -tearoff 1]
    $m add command -label "X axis" -command { Mirror_Piece  X }
    $m add command -label "Y axis" -command { Mirror_Piece  Y }
    pack .piece_edit.topmenu.mirror -side left

    menubutton .piece_edit.topmenu.color -text "Color" \
	-menu .piece_edit.topmenu.color.menu \
	-underline 0
    set m [menu .piece_edit.topmenu.color.menu -tearoff 1]
    $m add command -label "Browse Names" -command { Browse_Colors } \
	-underline 0
    $m add command -label "Choose in Palette" -command { Choose_Colors } \
	-underline 0
    $m add command -label "Pick at random" -command { Get_Random_Color } \
	-underline 0

    pack .piece_edit.topmenu.color -side left

    # making the "attributes" panel
    array set cur_piece_props [array get default_piece_props]
    array set old_props [array get cur_piece_props]

    set att [frame .piece_edit.att]
    label $att.label -text "Attributes of the current piece"

    frame $att.number
    label $att.number.label -text "Number" -width 15 -anchor w
    label $att.number.field \
	-textvar cur_piece_props(id) -width 10 
    frame $att.weight
    label $att.weight.label -text "Weight" -width 15 -anchor w
    entry $att.weight.field \
	-textvar cur_piece_props(weight) -width 10 \
	-relief sunken
    frame $att.color
    label $att.color.label -text "Color" -width 15 -anchor w
    entry $att.color.field \
	-textvar cur_piece_props(color) -width 10 \
	-relief sunken
    label $att.color.sample -text "" -width 2 -anchor w \
	-background $cur_piece_props(color)

    pack $att.label
    pack $att.number.label $att.number.field -side left
    pack $att.weight.label $att.weight.field -side left
    pack $att.color.label $att.color.field $att.color.sample -side left
    pack $att.number $att.weight $att.color -side top -fill x
    pack $att

    bind $att.color.field <Leave> "if \[catch {$att.color.sample configure -background \[$att.color.field get\]\n Recolor_Current_Piece}\] { set cur_piece_props(color) \$default_piece_props(color)}"

    # making the editing buttons
    set com [frame .piece_edit.command]
    frame $com.move -width 72 -height 72
    foreach dir { nw n ne w nil e sw s se } {
	set img [image create photo -format gif -file "$img_dir/$dir.gif"]
	button $com.move.$dir -image $img \
	    -highlightthickness 0 \
	    -command "Move_Piece $DX($dir) $DY($dir)"	
	place $com.move.$dir -in $com.move \
	    -x [expr 24*(1+$DX($dir))] \
	    -y [expr 24*(1-$DY($dir))]
    }
    pack $com.move -side top

    frame $com.rot 
    set img [image create photo -format gif -file "$img_dir/rot_plus.gif"]
    button $com.rot.plus -image $img -command "Rotate_Piece 1" \
	    -highlightthickness 0
    set img [image create photo -format gif -file "$img_dir/rot_minus.gif"]
    button $com.rot.minus -image $img -command "Rotate_Piece 3" \
	    -highlightthickness 0
    pack $com.rot.plus $com.rot.minus -side left
    pack $com.rot -side top

    frame $com.mirror 
    set img [image create photo -format gif -file "$img_dir/mirror_x.gif"]
    button $com.mirror.x -image $img -command "Mirror_Piece X" \
	    -highlightthickness 0
    set img [image create photo -format gif -file "$img_dir/mirror_y.gif"]
    button $com.mirror.y -image $img -command "Mirror_Piece Y" \
	    -highlightthickness 0
    pack $com.mirror.x $com.mirror.y -side left
    pack $com.mirror -side top

    pack $com -side left -fill y

    # making the edit frame (.*.graph)
    set nbcols $cur_width
    set width [expr $nbcols*$sq_size(.graph)]
    set height [expr $nbrows(.graph)*$sq_size(.graph)]
    set visible_height [expr $height>600?600:$height]

    if $DEBUG {
	puts "VH = $visible_height"
    }

    set g .piece_edit.graph

    if [winfo exists $g] {
	$g.canvas configure -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height]
    } else {
	frame $g
	canvas $g.canvas -background lightblue \
	    -width $width -height $visible_height \
	    -scrollregion [list 0 0 $width $height] \
	    -yscrollcommand [list $g.yscroll set]
	scrollbar $g.yscroll -orient vertical -width 10 \
	    -command [list $g.canvas yview]
    }

    pack $g.canvas -side left -fill y -expand true
    pack $g.yscroll -side right -fill y
    pack $g -side right

    # draw the grid
    for { set i 0 } { $i <= $nbcols } { incr i 1 } {
	set x [expr $i*$sq_size(.graph)]
	if $DEBUG {
	    puts "H  $x 0 $x [expr $nbrows(.graph)*$sq_size(.graph)]"
	}
	$g.canvas create line $x 0 $x [expr $nbrows(.graph)*$sq_size(.graph)] \
	    -tag {SEP HORR}
    }
    for { set i 0 } { $i <= $nbrows(.graph) } { incr i 1 } {
	set y [expr $i*$sq_size(.graph)]
	$g.canvas create line 0 $y [expr $nbcols*$sq_size(.graph)] $y \
	    -tag {SEP HORR}
    }

    puts [$g.canvas find all]
    # set viewpoint at bottom of canvas
    $g.canvas yview moveto 1.0

    # now standard bindings
    $g.canvas configure -cursor "@$img_dir/piece.xbm black"
    bind $g.canvas <1> "Set_Current_Square $g \[%W canvasx %x\] \[%W canvasy %y\] "
    bind $g.canvas <ButtonRelease-1> "Put_Square $g \[%W canvasx %x\] \[%W canvasy %y\] "
    bind $g.canvas <B1-Motion> "Move_Current_Square $g \[%W canvasx %x\] \[%W canvasy %y\] "
    bind $g.canvas <KeyPress-q> { Quit_Piece_Editor }
    bind $g.canvas <KeyPress-s> { Store_Current_Piece 0 0 }
    bind $g.canvas <Return> { Store_Current_Piece 0 1 }
    bind $g.canvas <KeyPress-1> { Move_Piece -1 -1 }
    bind $g.canvas <KeyPress-2> { Move_Piece  0 -1 }
    bind $g.canvas <KeyPress-3> { Move_Piece  1 -1 }
    bind $g.canvas <KeyPress-4> { Move_Piece -1  0 }
    bind $g.canvas <KeyPress-6> { Move_Piece  1  0 }
    bind $g.canvas <KeyPress-7> { Move_Piece -1  1 }
    bind $g.canvas <KeyPress-8> { Move_Piece  0  1 }
    bind $g.canvas <KeyPress-9> { Move_Piece  1  1 }
    bind $g.canvas <KeyPress-plus>  { Rotate_Piece  1 }
    bind $g.canvas <KeyPress-minus> { Rotate_Piece  3 }
    bind $g.canvas <KeyPress-x> { Mirror_Piece  X }
    bind $g.canvas <KeyPress-y> { Mirrir_Piece  Y }

    # prepare new piece...
    New_Piece
    
    # position the window
    # wm geometry .piece.edit "+[expr $Xpos+$object_distance/2]+$Ypos"

    # focus $g.canvas
}

proc Quit_Piece_Editor { } {
    global state

    if { "$state(.graph)" != "saved" } {
	switch [tk_dialog .quit "Modified" "The piece has been modified. Do you wish to store it?" {warning} 0 {Store (current) & Quit} {Store (new) & Quit} {Quit} {Cancel}] {
	    0 { 
		Store_Current_Piece 0 0
	    }
	    1 { 
		Store_Current_Piece 1 0
	    }
	    2 { }
	    3 { return }
	}
    }

    destroy .piece_edit

}

proc Recolor_Current_Piece { } {
    global cur_piece_props cur_piece

    foreach truc $cur_piece {
	set id [lindex $truc 0]
	.piece_edit.graph.canvas itemconfigure $id \
	    -fill $cur_piece_props(color)
    }
    
}

# basic procedure for drawing the square in the determined canvas
# (drop or graph) at a determined _GRID_ location
proc Draw_Square { graph i j color } {
    global sq_size ison nbrows

    set xpos [expr $i*$sq_size($graph) ]
    set ypos [expr ($nbrows($graph)-$j)*$sq_size($graph) ]
    
    set res [$graph.canvas create rectangle \
		 $xpos $ypos \
		 [expr $xpos+$sq_size($graph)] \
		 [expr $ypos-$sq_size($graph)] \
		 -fill $color -tag PIECE]

    return $res
}

proc Draw_Store_Square { i j color } {
    global sq_size nbrows store_ratio 

    set xpos [expr $i*$sq_size(.store)]
    set ypos [expr ($nbrows(.store)-$j)*$sq_size(.store)]
    .store.canvas create rectangle \
	$xpos $ypos \
	[expr $xpos+$sq_size(.store)] [expr $ypos-$sq_size(.store)] \
	-fill $color -tag PIECE
}

# toggles the square pointed at in the graph (canvas) g
proc Put_Square { graph x y } {
    global sq_size cur_piece cur_piece_props ison nbrows state
    global DEBUG

    # compute grid position
    set i [expr int($x) / $sq_size(.graph) ]
    set j [expr ($nbrows(.graph)*$sq_size(.graph) - int($y)) \
	       / $sq_size(.graph) ]

    append id "l" $i "c" $j
    if $DEBUG {
	puts "square id $id"
    }

    if { ![info exists ison($id)] } {
	set ison($id) [Draw_Square $graph $i $j $cur_piece_props(color)]
	# add square to current piece
	set cur_piece [lappend cur_piece [list $ison($id) $i $j]]

	if $DEBUG {
	    puts "CP=$cur_piece"
	}
    } else {
	$graph.canvas delete $ison($id)
	# remove square from list
	set target [lsearch -glob $cur_piece "* $i $j"]
	set cur_piece [lreplace $cur_piece $target $target ""]
	if $DEBUG {
	    puts "removing item $target gives $cur_piece"
	}

	unset ison($id)
    }

    set state(.graph) modified
}

######################################################################
# Display_Piece -
# Drawing (and loading) a piece specified by lo/hi profile.
#
# Warning: piece_id is not necessarily cur_piece_id, in order to be
# able to create copies of pieces.
#
# Parameters:
# - piece_id	the id of the piece in the piece store
#
proc Display_Piece { piece_id } { 
    global cur_width cur_piece cur_piece_props ison Piece_Store cur_piece_id
    global color weight
    global DEBUG

    if $DEBUG {
	puts "displaying piece number $piece_id"
    }

    set piece $Piece_Store($piece_id)
    array set lo_prof [lindex $piece 0]
    array set hi_prof [lindex $piece 1]
    if $DEBUG {
	puts "drawing piece [array get lo_prof] / [array get hi_prof]"
    }
    Empty_Graph .piece_edit.graph

    set cur_piece {}
    set cur_piece_props(id)     $cur_piece_id
    set cur_piece_props(weight) $weight($piece_id)
    set cur_piece_props(color)  $color($piece_id)

    # recoloring color sample
    catch [.piece_edit.att.color.sample configure \
	       -background $cur_piece_props(color)]

    if [array exists ison] { unset ison }

    for { set i 0 } { $i < $cur_width } { incr i 1 } {
	if { [string compare $lo_prof($i) eps ] != 0 } {
	    for { set j $lo_prof($i) } { $j < $hi_prof($i) } { incr j 1 } {
		set id ""
		append id "l" $i "c" $j
		set ison($id) [Draw_Square .piece_edit.graph \
				   $i $j $cur_piece_props(color)]
		set cur_piece [lappend cur_piece [list $ison($id) $i $j]]
	    }
	}
    }
}

######################################################################
# Display_Store -
# (re) drawing all the store in the interval [from,to], measured
# in piece numbers. The "to" field may be the keyword "end".
# The vertical separator is re-drawn completely in the process.
#
# Parameters:
# - from	the number of the piece from which to draw
# - to		the number of the piece to which to draw
#
proc Display_Store { from to } {
    global cur_piece_nb Piece_Store store_level weight color
    global sq_size cur_width nbrows

    if { 0 == [string compare $to end] } {
	set to $cur_piece_nb
    }
    set $to [expr $to > $cur_piece_nb ? $cur_piece_nb : $to]

    for { set k $from } { $k <= $to } { incr k 1 } {
	set level [expr 1+$store_level([expr $k-1])]

	array set lo_prof [lindex $Piece_Store($k) 0]
	array set hi_prof [lindex $Piece_Store($k) 1]

	for { set i 0 } { $i < $cur_width } { incr i 1 } {
	    if { [string compare $lo_prof($i) eps ] != 0 } {
		if { 0 == [string compare $hi_prof($i) eps] } {
		    ReportError "Store Error" \
			"Incoherent high profile in piece $k col $i"
		} elseif { $lo_prof($i) > $hi_prof($i) } {
		    ReportWarning "Store Error" \
			"Hi/Lo profiles incompatible\nin piece $k col $i\n($lo_prof($i) > $hi_prof($i))"
		} else {
		    for { set j $lo_prof($i) } { $j < $hi_prof($i) } \
			{ incr j 1 } {
			Draw_Store_Square $i [expr $j+$level] $color($k)
		    }
		}
	    }
	}
	# write number
	set x [expr ($cur_width+1)*$sq_size(.store)]
	set y [expr $sq_size(.store)*($nbrows(.store)-($store_level([expr $k-1])+$store_level($k))/2)]
	.store.canvas create text $x $y -text "$k" -tag TEXT

	# horizontal separator
	set level $store_level($k)
	set y [expr ($nbrows(.store)-$level)*$sq_size(.store) ]
	.store.canvas create line 0 $y [expr $cur_width*$sq_size(.store)] $y \
	    -tag {SEP HORR}
    }
    # Vertical separator. The entire separator is re-drawn each time
    # in order to avoid a multitude of broken lines.
    .store.canvas delete withtag VERT
    set x [expr $cur_width*$sq_size(.store) ]
    set yfrom [expr ($nbrows(.store)-$store_level($cur_piece_nb))*$sq_size(.store) ]
    set yto [expr $nbrows(.store)*$sq_size(.store) ]
    # puts "VL \{$x $yfrom -> $x $yto\}"
    .store.canvas create line $x $yfrom $x $yto -tag {SEP VERT}

}

proc Highlight_Store { from to mode } {
    global cur_piece_nb Piece_Store store_level weight color
    global sq_size cur_width nbrows

    if { [string compare $to end] == 0 } {
	set $to $cur_piece_nb
    } else {
	set $to [expr $to > $cur_piece_nb ? $cur_piece_nb : $to]
    }

    if { 0 == [string compare $mode on] } {
	set level $store_level([expr $from-1])
	set y1 [expr ($nbrows(.store)-$level)*$sq_size(.store)]
	set level $store_level($to)
	set y2 [expr ($nbrows(.store)-$level)*$sq_size(.store)]

	set hil [.store.canvas create rectangle \
		     0 $y1 [expr $cur_width*$sq_size(.store)] $y2 \
		     -fill white -tag HIGHLIGHT]
	# put the highlight item at the bottom of the list
	set bottom [lindex [.store.canvas find all] 0]
	.store.canvas lower $hil $bottom
    } else {
	.store.canvas delete withtag HIGHLIGHT
    }
}

######################################################################
# Expand_Store - 
# Check that store has enough levels in altitude. If not, expand the canvas
# towards the top.
# 
# Parameters:
# - level	the altitude to reach
#
proc Expand_Store { level } {
    global nbrows sq_size
    global DEBUG
    
    if $DEBUG {
	puts "Expanding store to $level"
    }
    if [expr $nbrows(.store) >= $level ] { return }

    set box [lindex [.store.canvas configure -scrollregion] 4]
    if $DEBUG {
	puts "... it needed it ($box)"
    }
    set box [lreplace $box 3 3 [expr $level*$sq_size(.store)]]
    .store.canvas configure -scrollregion $box
    set nbrows(.store) $level

    # redraw it completely
    Clear_Store 0 end
    Display_Store 1 end
}

######################################################################
# Set_Store_Width - 
# Sets the width of the store. When the store exists already, one must
# tell by how much the store must be expanded on both sides.
# The existing pieces are truncated if necessary. No questions are asked:
# the function Confirm_Truncation must be called before if necessary.
#
# Parameters:
# - left	by how much to expand on the left. If negative: shrink
# - right	by how much to expand on the right. If negative: shrink
#
proc Set_Store_Width { left right } {
    global cur_width cur_piece_nb Piece_Store 

    # eliminate degenerate and impossible operations
    if [expr { $left == 0 } && { $right == 0 } ] { return }
    set new_width [expr $cur_width + $left + $right]
    if { $new_width <= 0 } {
	ReportError "Store Error" "Cannot set width to $new_width."
	return
    }

    # Reshape store display. It has to be done here, otherwise the call
    # to "Is_Empty_Piece" below will fail. Also, the store must be cleared
    # before the "Make_Canvas" changes its size.
    set old_width $cur_width
    set cur_width $new_width
    Clear_Store 0 end
    Make_Canvas

    # Reshape pieces. In the process, some of them may disappear!
    set new_piece_nb 0
    for { set p 1 } { $p <= $cur_piece_nb } { incr p 1 } {
	# puts "Changing piece $p"
	set piece $Piece_Store($p)
	array set lo_prof [lindex $piece 0]
	array set hi_prof [lindex $piece 1]

	for { set i 0 } { $i < $new_width } { incr i 1 } {
	    set j [expr $i - $left]
	    # puts "NP $p, $i -> $j"
	    if [expr { $j >= 0 } && { $j < $old_width }] {
		set new_lo_prof($i) $lo_prof($j)
		set new_hi_prof($i) $hi_prof($j)
	    } else {
		set new_lo_prof($i) eps
		set new_hi_prof($i) eps
	    }
	}

	set new_piece [list [array get new_lo_prof] [array get new_hi_prof]]
	# now check if piece is empty.
	if ![Is_Empty_Piece $new_piece] {
	    incr new_piece_nb
	    set Piece_Store($new_piece_nb) $new_piece
	}
    }

    set cur_piece_nb $new_piece_nb

    Display_Store 1 end

    set state(.store) modified
}

######################################################################
# Make_Store_Editor -
# Create a dialog window for setting the width of the store or other
# attributes, sich as its square width.
# Handles in particular the case where the store already exists so that
# it must be widened or shrank.
# Returns the top window name in order to be used with tkwait for instance.
# 
# Parameters: none
#
proc Make_Store_Editor { } {
    global d_left d_right cur_width new_width sq_size
    global DEBUG

    if [catch {toplevel .store_edit}] { return }

    wm title .store_edit "Store Attribute Editor"

    # save old values in case of "cancel". Sets the "deltas" to neutral.
    set d_left 0
    set d_right 0
    set new_width $cur_width

    # making the "attributes" panel
    frame .store_edit.cols
    label .store_edit.cols.lab -text "Number of columns: "
    label .store_edit.cols.curlab -text "current: "
    label .store_edit.cols.curval -width 4 -text "$cur_width" 
    label .store_edit.cols.newlab -text "new: " 
    label .store_edit.cols.newval -width 4 -textvar new_width

    pack .store_edit.cols.lab \
	.store_edit.cols.curlab .store_edit.cols.curval \
	.store_edit.cols.newlab .store_edit.cols.newval \
	-side left
    pack .store_edit.cols -side top

    # modification area
    set recompute "set new_width \[expr \$cur_width+\$d_left+\$d_right\]"
    set f [frame .store_edit.modif]
    label $f.lab -text "Widen: "
    foreach dir { left right } {
	label $f.$dir -text "to the $dir"
	button $f.i_$dir -text "+" -command "incr d_$dir 1\n$recompute" \
	    -padx 3 -pady 0
	button $f.d_$dir -text "-" -command "incr d_$dir -1\n$recompute" \
	    -padx 3 -pady 0
	entry $f.val$dir -width 4 -textvar d_$dir
	pack $f.$dir $f.i_$dir $f.d_$dir $f.val$dir -side left -anchor w
	bind $f.val$dir <Leave> $recompute
	bind $f.i_$dir <2> "incr d_$dir 5\n$recompute"
	bind $f.d_$dir <2> "incr d_$dir -5\n$recompute"
	bind $f.i_$dir <3> "incr d_$dir 10\n$recompute"
	bind $f.d_$dir <3> "incr d_$dir -10\n$recompute"
    }
    pack $f -side top

    # square size
#     set f [frame .store_edit.sqsize]
#     label $f.lab -text "Square size"
#     entry $f.val -width 4 -textvar sq_size(.store)
#     pack $f.lab $f.val -side left -anchor w
#     pack $f -side top

    # buttons
    frame .store_edit.actions
    button .store_edit.actions.ok -text "Set" \
	-command "destroy .store_edit\n if { \$sq_size(.store) != $sq_size(.store) } { Make_Canvas \$sq_size(.store) }\n Set_Store_Width \$d_left \$d_right"
    button .store_edit.actions.cancel -text "Cancel" \
	-command "destroy .store_edit\n set sq_size(.store) $sq_size(.store)\n set exitcode cancel"
    pack .store_edit.actions.ok .store_edit.actions.cancel -side left
    pack .store_edit.actions -side top

    bind .store_edit <Return> {.store_edit.actions.ok invoke}
    bind .store_edit <Escape> {.store_edit.actions.cancel invoke}

    # return window name
    return .store_edit
}

######################################################################
# Store_Piece -
# Storing in the storage band the new piece, in the normalized format
# low/high profiles.
# The boolean as_new controls the fact that the piece is stored instead
# of its current version (in the case it was taken initially from the store)
# or as a new piece.
#
# Parameters
# - newpiece	the piece, as a pair of profiles
# - as_new	boolean
#
proc Store_Piece { newpiece as_new } {
    global cur_piece_props
    global cur_piece_nb sq_size nbrows cur_widt
    global Piece_Store store_level weight color piece_height state
    global DEBUG

    if { $DEBUG } {
	puts -nonewline "storing piece [lindex $newpiece 0] / [lindex $newpiece 1]"
	if $as_new {
	    puts " as new"
	} else {
	    puts " as current (nb $cur_piece_props(id)$)"
	}
    }

    # determining the id of the piece in the store 
    if $as_new {
	set cur_piece_props(id) [expr $cur_piece_nb + 1]
    }

    # putting the piece in the store, updating the total number if needed
    set Piece_Store($cur_piece_props(id)) $newpiece
    if { $cur_piece_props(id) > $cur_piece_nb } {
	set cur_piece_nb $cur_piece_props(id)
    }

    # setting values to current ones
    set color($cur_piece_props(id))  $cur_piece_props(color)
    set weight($cur_piece_props(id)) $cur_piece_props(weight)

    set piece_height($cur_piece_props(id)) [high_point $newpiece]
    if $DEBUG {
	puts "H($cur_piece_props(id))=piece_height($cur_piece_props(id))"
    }
    set oldbottom $store_level([expr $cur_piece_props(id)-1])
    if [info exists store_level($cur_piece_props(id))] {
	set oldtop $store_level($cur_piece_props(id))
    } else {
	set oldtop sky
    }
    set store_level($cur_piece_props(id)) \
	[expr $store_level([expr $cur_piece_props(id)-1]) \
	     + $piece_height($cur_piece_props(id)) + 2 ]

    if [expr { $oldbottom != $store_level([expr $cur_piece_props(id)-1]) } \
	    || { $oldtop != $store_level($cur_piece_props(id)) }] {
	# (re) computing store levels
	for { set i $cur_piece_props(id) } { $i <= $cur_piece_nb } { incr i 1 } {
	    set store_level($i) [expr $store_level([expr $i-1]) \
				     + $piece_height($i) + 2 ]
	}

	# displaying the piece and the rest of the store 
	Clear_Store $store_level([expr $cur_piece_props(id)-1]) end
	Display_Store $cur_piece_props(id) $cur_piece_nb
    } else {
	# just redisplaying the piece
	Clear_Store $store_level([expr $cur_piece_props(id)-1]) \
	    $store_level($cur_piece_props(id))
	Display_Store $cur_piece_props(id) $cur_piece_props(id)
    }

    set state(.store) modified
    set state(.graph) saved

    if $DEBUG {
	puts -nonewline "Storage area of piece $cur_piece_props(id) "
	puts "= \[$store_level([expr $cur_piece_props(id)-1]) $store_level($cur_piece_props(id))\["
    }

    # expanding store (will be done only if necessary)
    Expand_Store $store_level($cur_piece_props(id))
}

proc Store_Current_Piece { as_new clear } {
    global cur_piece cur_piece_props

    Store_Piece [Normalize_Piece $cur_piece] $as_new

    # highlighting off then on is necessary because of the possible change
    # in height!
    Highlight_Store 1 end off

    if { $clear } {
	New_Piece
    } else {
	Highlight_Store $cur_piece_props(id) $cur_piece_props(id) on
    }

}

#
# Delete a piece and pack the store
#
proc Delete_Piece { ypos } {
    global Piece_Store store_level color weight piece_height cur_piece_nb

    set number [Find_Piece $ypos]

    if { 0 < $number } {
	puts -nonewline "Deleting piece = $number"
	for { set i $number } { $i < $cur_piece_nb } { incr i 1 } {
	    set Piece_Store($i) $Piece_Store([expr $i+1])
	    set store_level($i) \
		[expr $store_level([expr $i-1]) + $store_level([expr $i+1]) \
		     - $store_level($i) ]
	    set weight($i) $weight([expr $i+1])
	    set color($i) $color([expr $i+1])
	    set piece-height($i) $piece_height([expr $i+1])
	}
	incr cur_piece_nb -1
	puts "... $cur_piece_nb remaining."
	Clear_Store $store_level([expr $number-1]) end
	Display_Store $number $cur_piece_nb
    } else {
	puts "none found"
    }
}

# procedure Delete_Square is obsolete: deleting a square is done by
# "toggling" it.
proc Delete_Square { x y } {
    global ison nbrows sq_size

    # compute grid position
    set i [expr $x / $sq_size(.graph) ]
    set j [expr ($nbrows(.graph)*$sq_size(.graph) - $y) / $sq_size(.graph) ]

    append id "l" $i "c" $j
    if [info exists ison($id)] {
	Put_Square .graph $x $y
    }
}

proc Set_Current_Square { graph xto yto } {
    global cur_sq_x cur_sq_y sq_size nbrows

    set cur_sq_x [expr int($xto) / $sq_size(.graph) ]
    set cur_sq_y [expr ($nbrows(.graph)*$sq_size(.graph) - int($yto)) \
	       / $sq_size(.graph) ]
}
    
proc Move_Current_Square { graph xto yto } {
    global cur_sq_x cur_sq_y sq_size nbrows
    global DEBUG

    if $DEBUG {
	puts "mv $xto $yto"
    }
    # compute grid position
    set i [expr int($xto) / $sq_size(.graph) ]
    set j [expr ($nbrows(.graph)*$sq_size(.graph) - int($yto)) \
	       / $sq_size(.graph) ]

    if [expr { $i != $cur_sq_x } || { $j != $cur_sq_y }] {
	if $DEBUG {
	    puts "++ ($cur_sq_x,$cur_sq_y) ---> ($i,$j)"
	}
	Put_Square $graph $cur_sq_x $cur_sq_y
	set cur_sq_x $i
	set cur_sq_y $j
	Put_Square $graph $cur_sq_x $cur_sq_y
    }
}

proc Move_Piece { dx dy } {
    global ison nbrows sq_size cur_piece cur_width
    global DEBUG

    if $DEBUG {
	puts "DXDY $dx $dy"
    }

    set newpiece ""

    foreach square $cur_piece {
	if [expr { [llength $square ] == 3 }] {

	    set id [lindex $square 0]
	    set x  [lindex $square 1]
	    set y  [lindex $square 2]

	    incr x $dx
	    incr y $dy

	    if [expr { $x >= 0 } && { $x < $cur_width } \
		    && { $y >= 0 } && { $x < $nbrows(.piece_edit.graph) }] {
		set pos ""
		append pos "l" $x "c" $y
		set newon($pos) $id
		
		.piece_edit.graph.canvas move $id \
		    [expr $sq_size(.piece_edit.graph)*$dx] \
		    [expr -$sq_size(.piece_edit.graph)*$dy] 
		
		lappend newpiece [list $id $x $y]
	    } else {
		.piece_edit.graph.canvas delete $id 
	    }
	}
    } 

    unset ison
    array set ison [array get newon]
    set cur_piece $newpiece
}

#
# rotates the piece of "nb_right_angles" quarters, clockwise
#
proc Rotate_Piece { nb_right_angles } {
    global ison nbrows sq_size cur_piece cur_width
    global DEBUG

    # computing center of rotation
    set bb [Bounding_Box $cur_piece]
    set xcenter [expr ([lindex $bb 0]+[lindex $bb 2])/2.0]
    set ycenter [expr ([lindex $bb 1]+[lindex $bb 3])/2.0]

    # adjusting center to avoid non-integrality
    if { 0 != [expr int($xcenter+$ycenter)-($xcenter+$ycenter)] } {
	if { 0 != [expr int($xcenter)-$xcenter] } {
	    set xcenter [expr int($xcenter+1)]
	} else {
	    set ycenter [expr int($ycenter+1)]
	}
    }

    if $DEBUG {
	puts "ROT $nb_right_angles rights around ($xcenter,$ycenter) "
    }

    set newpiece ""
    
    foreach square $cur_piece {
	if [expr { [llength $square ] == 3 }] {
	    
	    set id [lindex $square 0]
	    set x0 [lindex $square 1]
	    set y0 [lindex $square 2]
	    set x $x0
	    set y $y0

	    for { set nb 0 } { $nb < $nb_right_angles } { incr nb 1 } {
		set i [expr  $xcenter+$ycenter-$y-1]
		set j [expr -$xcenter+$ycenter+$x]
		set x $i
		set y $j
	    }

	    set x [expr int($x)]
	    set y [expr int($y)]

	    if $DEBUG {
		puts "($x0,$y0)->($x,$y)"
	    }

	    if [expr { $x >= 0 } && { $x < $cur_width } \
		    && { $y >= 0 } && { $y < $nbrows(.piece_edit.graph) }] {
		set pos ""
		append pos "l" $x "c" $y
		set newon($pos) $id
		
		set dx [expr $x-$x0]
		set dy [expr $y-$y0]

		.piece_edit.graph.canvas move $id \
		    [expr  $sq_size(.piece_edit.graph)*$dx] \
		    [expr -$sq_size(.piece_edit.graph)*$dy] 
		
		lappend newpiece [list $id $x $y]

	    } else {
		.piece_edit.graph.canvas delete $id 
	    }
	}
    } 

    unset ison
    array set ison [array get newon]
    set cur_piece $newpiece
}

#
# Mirror the piece according to the specified axis (must be "X" or "Y")
#
proc Mirror_Piece { axis } {
    global ison nbrows sq_size cur_piece cur_width
    global DEBUG

    # computing center of the bounding box
    set bb [Bounding_Box $cur_piece]
    set xcenter_x_2 [expr [lindex $bb 0]+[lindex $bb 2]]
    set ycenter_x_2 [expr [lindex $bb 1]+[lindex $bb 3]]

    if $DEBUG {
	puts "MIR $axis"
    }

    set newpiece ""
    
    foreach square $cur_piece {
	if [expr { [llength $square ] == 3 }] {
	    
	    set id [lindex $square 0]
	    set x0 [lindex $square 1]
	    set y0 [lindex $square 2]

	    if { 0 == [string compare $axis X] } {
		set x [expr $xcenter_x_2-$x0-1]
		set y $y0
	    } else {
		set x $x0
		set y [expr $ycenter_x_2-$y0-1]
	    }

	    set pos ""
	    append pos "l" $x "c" $y
	    set newon($pos) $id
		
	    set dx [expr $x-$x0]
	    set dy [expr $y-$y0]

	    .piece_edit.graph.canvas move $id \
		[expr  $sq_size(.piece_edit.graph)*$dx] \
		[expr -$sq_size(.piece_edit.graph)*$dy] 
		
	    lappend newpiece [list $id $x $y]

	}
    }

    unset ison
    array set ison [array get newon]
    set cur_piece $newpiece
}
